package ru.ulstu.fc.rule.service;

import com.fuzzylite.Engine;
import com.fuzzylite.activation.General;
import com.fuzzylite.defuzzifier.WeightedAverage;
import com.fuzzylite.norm.s.Maximum;
import com.fuzzylite.norm.t.AlgebraicProduct;
import com.fuzzylite.norm.t.Minimum;
import com.fuzzylite.rule.Rule;
import com.fuzzylite.rule.RuleBlock;
import com.fuzzylite.term.Triangle;
import com.fuzzylite.variable.InputVariable;
import com.fuzzylite.variable.OutputVariable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import ru.ulstu.fc.project.service.ProjectRulesService;
import ru.ulstu.fc.project.service.ProjectService;
import ru.ulstu.fc.project.service.ProjectVariableService;
import ru.ulstu.fc.rule.model.FuzzyRule;
import ru.ulstu.fc.rule.model.FuzzyTerm;
import ru.ulstu.fc.rule.model.OutputValue;
import ru.ulstu.fc.rule.model.ProjectInferenceData;
import ru.ulstu.fc.rule.model.Variable;
import ru.ulstu.fc.rule.model.dto.VariableValueDto;

import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Service
public class FuzzyInferenceService {
    private final static Logger LOG = LoggerFactory.getLogger(FuzzyInferenceService.class);
    private final static String OUTPUT_VARIABLE_NAME = "кредит";
    private final static String RULE_TEMPLATE = "if %s is %s and %s is %s then "
            + OUTPUT_VARIABLE_NAME
            + " is %s";
    private final static String NO_RESULT = "Нет результата";
    private final Engine fuzzyEngine;
    private final ProjectService projectService;
    private final FuzzyRuleService fuzzyRuleService;
    private final ProjectRulesService projectRulesService;
    private final ProjectVariableService projectVariableService;

    public FuzzyInferenceService(Engine fuzzyEngine, ProjectService projectService, FuzzyRuleService fuzzyRuleService, ProjectRulesService projectRulesService, ProjectVariableService projectVariableService) {
        this.fuzzyEngine = fuzzyEngine;
        this.projectService = projectService;
        this.fuzzyRuleService = fuzzyRuleService;
        this.projectRulesService = projectRulesService;
        this.projectVariableService = projectVariableService;
    }

    private List<String> getDemoRules() {
        return Arrays.asList(
                String.format(RULE_TEMPLATE, "возраст", "молодой", "доход", "высокий", "средний"),
                String.format(RULE_TEMPLATE, "возраст", "средний", "доход", "высокий", "большой"),
                String.format(RULE_TEMPLATE, "возраст", "старый", "доход", "средний", "средний")
        );
    }

    private InputVariable getInputVariable(Variable variable) {
        final InputVariable input = new InputVariable();
        input.setName(variable.getName());
        input.setDescription("");
        variable.getFuzzyTerms().sort(Comparator.comparing(FuzzyTerm::getCrispValue));
        input.setRange(0, variable.getFuzzyTerms().getLast().getCrispValue());
        input.setEnabled(true);
        input.setLockValueInRange(false);
        double prev = 0;
        for (int i = 0; i < variable.getFuzzyTerms().size(); i++) {
            Triangle term = new Triangle(variable.getFuzzyTerms().get(i).getDescription(),
                    prev,
                    variable.getFuzzyTerms().get(i).getCrispValue(),
                    variable.getFuzzyTerms().get(i).getCrispValue() + variable.getFuzzyTerms().get(i).getCrispValue() - prev);
            prev = term.getVertexB();
            input.addTerm(term);
        }
        return input;
    }

    private <T extends Enum<T>> OutputVariable getOutputVariable(Variable variable) {
        final OutputVariable output = new OutputVariable();
        output.setName(variable.getName());
        output.setDescription("");
        output.setRange(0, variable.getFuzzyTerms().getLast().getCrispValue());
        output.setEnabled(true);
        output.setAggregation(new Maximum());
        output.setDefuzzifier(new WeightedAverage());
        output.setDefaultValue(Double.NaN);
        output.setLockValueInRange(false);
        double prev = 0;
        for (int i = 0; i < variable.getFuzzyTerms().size(); i++) {
            Triangle term = new Triangle(
                    variable.getFuzzyTerms().get(i).getDescription(),
                    prev,
                    variable.getFuzzyTerms().get(i).getCrispValue(),
                    variable.getFuzzyTerms().get(i).getCrispValue() + variable.getFuzzyTerms().get(i).getCrispValue() - prev);
            prev = term.getVertexB();
            output.addTerm(term);
        }
        return output;
    }

    private RuleBlock getRuleBlock(Engine engine,
                                   List<String> rules,
                                   List<Variable> inputVariables,
                                   List<Variable> outputVariables) {
        inputVariables.stream().map(this::getInputVariable).forEach(engine::addInputVariable);
        outputVariables.stream().map(this::getOutputVariable).forEach(engine::addOutputVariable);

        RuleBlock mamdani = new RuleBlock();
        mamdani.setName("mamdani");
        mamdani.setDescription("");
        mamdani.setEnabled(true);
        mamdani.setConjunction(new Minimum());
        mamdani.setImplication(new AlgebraicProduct());
        mamdani.setActivation(new General());
        rules.forEach(r -> mamdani.addRule(Rule.parse(r, engine)));
        return mamdani;
    }

    private List<OutputValue> getConsequent(Engine engine, Map<String, Double> variableValues) {
        List<OutputVariable> outputVariables = engine.getOutputVariables();
        for (Map.Entry<String, Double> variableValue : variableValues.entrySet()) {
            InputVariable inputVariable = engine.getInputVariable(variableValue.getKey());
            inputVariable.setValue(variableValue.getValue());
        }
        engine.process();

        return outputVariables
                .stream()
                .filter(v -> !Double.isNaN(v.getValue()))
                .map(OutputVariable::fuzzyOutput)
                .map(a -> new OutputValue(a.getName(),
                        a.getTerms().getFirst().getTerm().getName(),
                        a.getTerms().getFirst().getDegree()))
                .toList();
    }

    public List<OutputValue> getFuzzyInference(Map<String, Double> vals) {
        return getFuzzyInference(getDemoRules(), vals,
                Arrays.asList(new Variable("возраст", Arrays.asList(
                                new FuzzyTerm("молодой", 35.0),
                                new FuzzyTerm("средний", 60.0),
                                new FuzzyTerm("старый", 100.0))
                        ),
                        new Variable("доход", Arrays.asList(
                                new FuzzyTerm("небольшой", 35000.0),
                                new FuzzyTerm("средний", 100000.0),
                                new FuzzyTerm("высокий", 500000.0))
                        )
                ),
                Arrays.asList(new Variable("кредит", Arrays.asList(
                        new FuzzyTerm("небольшой", 20000.0),
                        new FuzzyTerm("средний", 100000.0),
                        new FuzzyTerm("большой", 1000000.0))))
        );
    }

    public List<OutputValue> getFuzzyInference(List<String> rules,
                                               Map<String, Double> values,
                                               List<Variable> inputVariables,
                                               List<Variable> outputVariables) {
        fuzzyEngine.getRuleBlocks().clear();
        fuzzyEngine.getInputVariables().clear();
        fuzzyEngine.getOutputVariables().clear();
        fuzzyEngine.addRuleBlock(getRuleBlock(fuzzyEngine, rules, inputVariables, outputVariables));
        return getConsequent(fuzzyEngine, values);
    }

    public List<OutputValue> getProjectFuzzyInference(ProjectInferenceData projectInferenceData) {
        List<String> fuzzyRules = projectRulesService.getByProjectId(projectInferenceData.getProjectId())
                .stream()
                .map(FuzzyRule::getContent)
                .toList();
        Map<String, Double> variableValues = projectInferenceData.getVariableValues()
                .stream()
                .collect(Collectors.toMap(
                        VariableValueDto::getVariableName,
                        VariableValueDto::getValue));
        List<Variable> inputVariables = projectVariableService.getInputByProjectId(projectInferenceData.getProjectId());
        List<Variable> outputVariables = projectVariableService.getOutputByProjectId(projectInferenceData.getProjectId());

        return getFuzzyInference(fuzzyRules,
                variableValues,
                inputVariables,
                outputVariables);
    }

    public List<OutputValue> getProjectFuzzyInference(Integer projectId, Map<String, Double> variableValues) {
        List<String> fuzzyRules = projectRulesService.getByProjectId(projectId)
                .stream()
                .map(FuzzyRule::getContent)
                .toList();
        List<Variable> inputVariables = projectVariableService.getInputByProjectId(projectId);
        List<Variable> outputVariables = projectVariableService.getOutputByProjectId(projectId);

        return getFuzzyInference(fuzzyRules,
                variableValues,
                inputVariables,
                outputVariables);
    }
}
